// gcc -std=c99 -Wall -Wextra -pedantic -O3 coupled.c -lm

#include <math.h>
#include <stdio.h>

static const double a[6][5] = 
  { {0, 0, 0, 0, 0}
  , {1./4, 0, 0, 0, 0}
  , {3./32, 9./32, 0, 0, 0}
  , {1932./2197, -7200./2197, 7296./2197, 0, 0}
  , {439./216, -8, 3680./513, -845./4104, 0}
  , {-8./27, 2, -3544./2565, 1859./4104, -11./40}
  };

static const double b[2][6] =
  { {25./216, 0, 1408./2565, 2197./4104, -1./5, 0}
  , {16./135, 0, 6656./12825, 28561./56430, -9./50, 2./55}
  };

static inline void f(double *out, const double v[4])
{
  double out0 = v[2];
  double out1 = v[3];
  double out2 = -exp(v[1] * v[1]) * v[0];
  double out3 = -exp(v[0] * v[0]) * v[1];
  out[0] = out0;
  out[1] = out1;
  out[2] = out2;
  out[3] = out3;
}

static inline double E(const double v[4])
{
  double w0 = sqrt(exp(v[1] * v[1]));
  double w1 = sqrt(exp(v[0] * v[0]));
  double k0 = 0.5 * v[2] * v[2];
  double k1 = 0.5 * v[3] * v[3];
  double u0 = 0.5 * w0 * v[0] * v[0];
  double u1 = 0.5 * w1 * v[1] * v[1];
  return k0 + k1 + u0 + u1;
}

int rk4(double *v, double *hh)
{
  const double maxstep = 1;
  const double abstol = 1e-4;
  const double reltol = 1e-4;
  const double eps = 2.220446049250313e-16;

  double h = *hh;

  double k[6][4];
  f(k[0], v);
  for (int p = 1; p < 6; ++p)
  {
    double u[4] = { 0, 0, 0, 0 };
    for (int q = 0; q < p; ++q)
      for (int j = 0; j < 4; ++j)
        u[j] += a[p][q] * k[q][j];
    for (int j = 0; j < 4; ++j)
      u[j] = v[j] + h * u[j];
    f(k[p], u);
  }
  double u4[4] = { 0, 0, 0, 0 };
  for (int q = 0; q < 6; ++q)
    for (int j = 0; j < 4; ++j)
      u4[j] += b[0][q] * k[q][j];
  double u5[4] = { 0, 0, 0, 0 };
  for (int q = 0; q < 6; ++q)
    for (int j = 0; j < 4; ++j)
      u5[j] += b[1][q] * k[q][j];

  double norm = 0;
  for (int j = 0; j < 4; ++j)
    norm += fabs(v[j]);
  double delta = 0;
  for (int j = 0; j < 4; ++j)
    delta += fabs(u5[j] - u4[j]);
  delta = fmax(delta, eps);
  double tau = fmax(reltol * fmax(norm, 1), abstol);

  if (delta <= tau)
    for (int j = 0; j < 4; ++j)
      v[j] += h * u5[j];

  *hh = fmin(maxstep, 0.8 * h * pow(tau / delta, 1./6));

  return delta <= tau;
}

int main()
{
  double t = 0;
  double h = 0.01;
  double v[4] = { 0, 1, 1, 1 };
  while (t < 45000)
    if (rk4(v, &h))
    {
      t += h;
      printf("%f %f %f %f %f %f\n", t, E(v), v[0], v[1], v[2], v[3]);
    }
  return 0;
}
